import numpy as np
from scipy import ndimage

from lentil.convolvable import Pixel
from lentil.plane import Detector
from lentil.radiometry import Spectrum
from lentil import detector
from lentil import util

__all__ = ['FPA', 'BayerFPA']


class FPA(Detector):
    """Focal plane array base class.

    The entire signal flow through the FPA model is depicted in the figure
    below. Boxes represented by attributes which are ``None`` are simply passed
    through without modifying the signal in any way.

    .. image:: ../_static/img/detector_model.png

    Parameters
    ----------
    pixelscale : float
        Pixel size in meters. Pixels are assumed to be square.

    shape : int or array_like
        Number of pixels as (rows, cols). If a single value is provided, the
        :class:`FPA` is assumed to be square with nrows = ncols = shape.

    **kwargs : :class:`FPA` properties, optional
        Additional detector attributes which can be optionally specified. Any
        attributes not defined will default to None and have no effect on
        image simulations. The full list of supported attributes is:

        ======================== ======================================================================
        :attr:`charge_diffusion` :class:`~detector.ChargeDiffusion` model
        :attr:`cosmic_rays`      :class:`~detector.CosmicRay` model
        :attr:`dark_shot_noise`  Dark current :class:`~detector.ShotNoise` model
        :attr:`dark_signal`      :class:`~detector.DarkSignal` model
        :attr:`defects`          :class:`~detector.Defect` or list of :class:`~detector.Defect` objects
        :attr:`dsnu`             Dark signal nonuniformity :class:`~detector.FPN` model
        :attr:`gain`             :class:`~detector.Gain` model
        :attr:`offset_fpn`       Offset :class:`~detector.FPN` model
        :attr:`prnu`             Light signal nonuniformity :class:`~detector.FPN` model
        :attr:`qe`               QE as a :class:`~lentil.radiometry.Spectrum` or a scalar
        :attr:`read_noise`       :class:`~detector.ReadNoise` model
        :attr:`shot_noise`       :class:`~detector.ShotNoise` model
        ======================== ======================================================================

    Examples
    --------
    Additional arguments can be passed in individually:

    .. code:: pycon

        >>> f = FPA(pixelscale=5e-6, shape=(256, 256), read_noise=ReadNoise(50))

    They can also be packed into a dictionary and passed in that way:

    .. code:: pycon

        >>> f_attrs = {'shot_noise': ShotNoise(), 'dark_current': DarkCurrent(500)}
        >>> f = FPA(pixelscale=5e-6, shape=(256, 256), **f_attrs)

    """
    # TODO: rewrite this like Plane constructor. Also need to define __init_subclass__
    def __init__(self, pixelscale, shape, **kwargs):
        super().__init__(pixelscale, shape)

        self.qe = kwargs.get('qe', None)
        self.shot_noise = kwargs.get('shot_noise', None)
        self.prnu = kwargs.get('prnu', None)
        self.dark_signal = kwargs.get('dark_signal', None)
        self.dark_shot_noise = kwargs.get('dark_shot_noise', None)
        self.dsnu = kwargs.get('dsnu', None)
        self._gain = kwargs.get('gain', None)
        self._read_noise = kwargs.get('read_noise', None)
        self._offset_fpn = kwargs.get('offset_fpn', None)
        self._charge_diffusion = kwargs.get('charge_diffusion', None)
        self._defects = kwargs.get('defects', None)
        self._cosmic_rays = kwargs.get('cosmic_rays', None)

    @property
    def qe(self):
        """Quantum efficiency curve used to convert detected photons to
        electrons.

        Returns
        -------
        qe :  :class:`~lentil.radiometry.Spectrum`

        """
        return self._qe

    @qe.setter
    def qe(self, value):
        if value is None:
            self._qe = None
        elif isinstance(value, Spectrum):
            self._qe = value
        else:
            raise AttributeError('qe must be a Spectrum object or None')

    @property
    def shot_noise(self):
        """Shot noise model applied to the light signal. If ``None`` (default),
        shot noise is not applied.

        Returns
        -------
        shot_noise : :class:`~lentil.detector.ShotNoise` or None, optional

        """
        return self._shot_noise

    @shot_noise.setter
    def shot_noise(self, value):
        if value is None:
            self._shot_noise = None
        elif isinstance(value, detector.ShotNoise):
            self._shot_noise = value
        else:
            raise AttributeError('shot_noise must be a ShotNoise object or None')

    @property
    def prnu(self):
        """Photo response non-uniformity model. If ``None`` (default), PRNU is
        not applied.

        Returns
        -------
        prnu : :class:`~lentil.detector.FPN` or None, optional

        """
        return self._prnu

    @prnu.setter
    def prnu(self, value):
        if value is None:
            self._prnu = None
        elif isinstance(value, detector.FPN):
            self._prnu = value
        else:
            raise AttributeError('prnu must be an FPN object or None')

    @property
    def dark_signal(self):
        """Dark signal model. If not ``None`` (default), dark current is not
        applied.

        Returns
        -------
        dark_signal : :class:`~lentil.detector.DarkSignal` or None, optional

        """
        return self._dark_signal

    @dark_signal.setter
    def dark_signal(self, value):
        if value is None:
            self._dark_signal = None
        elif isinstance(value, detector.DarkSignal):
            self._dark_signal = value
        else:
            raise AttributeError('dark_signal must be a DarkSignal object or None')

    @property
    def dark_shot_noise(self):
        """Shot noise model applied to the dark signal. If ``None`` (default),
        shot noise is not applied.

        Returns
        -------
        dark_shot_noise : :class:`~lentil.detector.ShotNoise` or None, optional

        """
        return self._dark_shot_noise

    @dark_shot_noise.setter
    def dark_shot_noise(self, value):
        if value is None:
            self._dark_shot_noise = None
        elif isinstance(value, detector.ShotNoise):
            self._dark_shot_noise = value
        else:
            raise AttributeError('dark_shot_noise must be a ShotNoise object or None')

    @property
    def dsnu(self):
        """Dark signal fixed-pattern noise (DSNU) model. Multiple models are
        combined when specified as a list. If ``None`` (default), DSNU is not
        applied.

        Returns
        -------
        dsnu : :class:`~lentil.detector.FPN`, list_like or None, optional

        """
        return self._dsnu

    @dsnu.setter
    def dsnu(self, value):
        if value is None:
            self._dsnu = None
        elif isinstance(value, detector.FPN):
            self._dsnu = value
        else:
            raise AttributeError('dsnu must be an FPN object or None')

    @property
    def gain(self):
        """Gain model used to convert electrons to DN

        Returns
        -------
        gain : :class:`~lentil.detector.Gain`

        """
        return self._gain

    @gain.setter
    def gain(self, value):
        if value is None:
            self._gain = None
        elif isinstance(value, detector.Gain):
            self._gain = value
        else:
            raise AttributeError('gain must be a Gain object or None')

    @property
    def read_noise(self):
        """Read noise model applied during the readout of the analog signal. If
        ``None`` (default), read noise is not applied.

        Returns
        -------
        read_noise : :class:`~lentil.detector.ReadNoise` or None, optional

        """
        return self._read_noise

    @read_noise.setter
    def read_noise(self, value):
        if value is None:
            self._read_noise = None
        elif isinstance(value, detector.ReadNoise):
            self._read_noise = value
        else:
            raise AttributeError('read_noise must be a ReadNoise object or None')

    @property
    def offset_fpn(self):
        """Offset fixed pattern noise model. Multiple models are applied
        serially when specified as a list. If ``None`` (default), offset FPN is
        not applied.

        Returns
        -------
        offset_fpn : :class:`~lentil.detector.FPN`, list_like, or None, optional

        """
        return self._offset_fpn

    @offset_fpn.setter
    def offset_fpn(self, value):
        if value is None:
            self._offset_fpn = None
        elif isinstance(value, detector.FPN):
            self._offset_fpn = value
        else:
            raise AttributeError('offset_fpn must be an FPN object or None')

    @property
    def charge_diffusion(self):
        """Charge diffusion model describing how charge leaks into adjacent
        pixels. If ``None`` (default), charge diffusion is not applied.

        Returns
        -------
        charge_diffusion : :class:`~lentil.detector.ChargeDiffusion` or None,
        optional

        """
        return self._charge_diffusion

    @charge_diffusion.setter
    def charge_diffusion(self, value):
        if value is None:
            self._charge_diffusion = None
        elif isinstance(value, detector.ChargeDiffusion):
            self._charge_diffusion = value
        else:
            raise AttributeError('charge_diffusion must be a ChargeDiffusion '
                                 'object or None')

    @property
    def defects(self):
        """Detector defect model. Multiple models are combined when specified as
        a list. If ``None`` (default), defects are not included.

        Returns
        -------
        defects : :class:`~lentil.detector.Defect`, list_like, or None, optional

        """
        return self._defects

    @defects.setter
    def defects(self, value):
        if value is None:
            self._defects = None
        elif isinstance(value, detector.Defect):
            self._defects = value
        elif all(isinstance(defect, detector.Defect) for defect in list(value)):
            self._defects = list(value)
        else:
            raise AttributeError('defects must be a Defect object, a list of '
                                 'Defect objects, or None')

    @property
    def cosmic_rays(self):
        """Cosmic ray model. If ``None`` (default), cosmic rays are not
        included.

        Returns
        -------
        cosmic_rays : :class:`~lentil.cosmic.CosmicRay` or None, optional

        """
        return self._cosmic_rays

    @cosmic_rays.setter
    def cosmic_rays(self, value):
        if value is None:
            self._cosmic_rays = None
        elif isinstance(value, detector.CosmicRay):
            self._cosmic_rays = value
        else:
            raise AttributeError('cosmic_rays must be a CosmicRay object or '
                                 'None')

    def frame(self, flux, ts, wave=None, waveunit='nm', oversample=1,
              pixelate=False, collect_charge=False, window=None,
              warn_saturate=False):
        """Simulate a frame.

        Parameters
        ----------
        flux : array_like
            The flux presented to the sensor. ``flux`` should be an (nwave, nrows,
            ncols) datacube with units of photons/s where the first dimension
            represents wavelength if ``collect_charge`` is True or an
            (nrows, ncols) array with units of electrons/s if
            ``collect_charge`` is False.

        ts : float
            Integration time (seconds)

        wave : array_like, optional
            Wavelengths corresponding to each slice in ``flux``. The length of
            ``wave`` must be equal to the number of samples ``n`` in ``flux``.
            If ``collect_charge = False``, ``wave`` is unused and can be
            ``None``.

        waveunit : str, optional
            Units of ``wave``. Defaults is ``nm``

        oversample : int, optional
            Oversampling factor present in ``flux``. Default is 1.

        pixelate : bool, optional
            If True, ``flux`` is convolved with the pixel MTF before rescaling
            the image to native detector sampling. Default is False.
            :func:`pixelate` should only be used if ``oversample`` >= 2

        collect_charge : bool, optional
            If False (default), the input ``flux`` is assumed to be in
            electrons/s. If True, the input flux is assumed to be a data cube of
            photons/s (where the third dimension is wavelength) so the charge
            collection process (photons -> electrons) is performed as a part of
            frame generation.

        window : array_like or None, optional
            Indices of focal plane array to return given as (r_start, r_end,
            c_start, c_end). This definition follows standard Numpy indexing.

        warn_saturate : bool, optional
            Raise a warning when pixels saturate. Default is False.

        Returns
        -------
        frame : ndarray
            Raw frame

        Notes
        -----
        In cases where dynamic noise (jitter, smear, etc.) is present and needs
        to be applied in an oversampled image space, the following approach to
        frame generation is recommended:

            * Compute the polychromatic image in oversampled space with
              ``flatten = False``
            * Compute the binned detector irradiance at the wavelengths present
              in the polychromatic image stack (see:
              :func:`~lentil.radiometry.Spectrum.bin`)
            * Call :func:`collect_charge` to simulate the charge collection
              process. This method will flatten the image, speeding up the
              application of convolutional dynamic noise models
            * Call :func:`frame` with ``collect_charge = False``. Note that in
              this case, the ``wave`` parameter is unused and can be ``None``.

        """

        # If collect charge is True, we need to make sure the user has supplied
        # a wave and waveunit
        if collect_charge:
            assert wave is not None
            assert waveunit is not None

        # accumulate the light signal
        light_e = self.light(flux, ts, wave, waveunit, oversample, pixelate,
                             collect_charge, window)

        # accumulate the dark signal
        dark_e = self.dark(light_e.shape, ts, window)

        # cosmic rays
        if self.cosmic_rays:
            cosmic_e = self.cosmic_rays(light_e.shape, self.pixelscale, ts)
        else:
            cosmic_e = 0

        # combine the signals
        e = light_e + dark_e + cosmic_e

        # apply charge diffusion
        if self.charge_diffusion:
            e = self.charge_diffusion(e)

        # apply offset fpn
        e = self.offset(e, window)

        # read out the frame
        frame = self.readout(e, window, warn_saturate)

        return frame

    def collect_charge(self, count, wave, waveunit='nm', *args, **kwargs):
        """Convert photon count (or flux) to electron count (or flux) by
        applying the detector's wavelength-dependent quantum efficiency.

        Parameters
        ----------
        count : array_like
            The photons presented to the sensor. Should have shape (nwave,
            nrows, ncols)

        wave : array_like
            Wavelengths corresponding to each slice in ``count``. The length of
            ``wave`` must be equal to the number of samples ``nwave`` in ``count``.

        waveunit : str, optional
            Units of ``wave``. Defaults is ``nm``

        Returns
        -------
        elec : ndarray
            Electron count or flux

        Notes
        -----
        The units of ``count`` don't really matter, as long as the user is aware
        that this method converts photons per whatever to electrons per
        whatever. Whatever is nothing for counts and seconds for flux.

        """
        count = np.asarray(count)
        if count.ndim == 2:
            count = count[np.newaxis, ...]

        qe = self.qe.sample(wave, waveunit=waveunit)
        if qe.ndim == 0:
            qe = qe[..., np.newaxis]
        return np.einsum('ijk,i->jk', count, qe)

    def bias(self, shape=None, window=None):
        """Simulate a bias (zero integration time) frame

        Parameters
        ----------
        shape : array_like or None, optional
            Shape to

        window : array_like or None, optional
            Indices of focal plane array to return given as (r_start, r_end,
            c_start, c_end). This definition follows standard Numpy indexing.

        Returns
        -------
        bias : ndarray

        Notes
        -----
        This method is equivalent to calling :func:`frame` with an integration
        time of 0.

        """
        if shape is None:
            shape = self.shape
        else:
            shape = tuple(np.asarray(shape))

        # we don't collect any charge so we just read out a dark frame with zero
        # integration time
        e = self.dark(shape, 0, window)

        # apply charge diffusion
        if self.charge_diffusion:
            e = self.charge_diffusion(e)

        # apply offset fpn
        e = self.offset(e, window)

        # read out the frame
        frame = self.readout(e, window, warn_saturate=False)

        return frame

    def light(self, flux, ts, wave, waveunit, oversample, pixelate, collect_charge,
              window):

        # integrate
        count = self.integrate(flux, ts)

        # photon to electron conversion - there are two possibilities here:
        # 1. count is a wavelength dependent cube so we call collect_charge to
        #    apply the QE and collapse the cube into a single frame
        # 2. count is already collapsed into a single frame
        if collect_charge:
            light_e = self.collect_charge(count, wave, waveunit, oversample, window)
        else:
            light_e = count

        # pixelate
        if pixelate:
            light_e = self.pixelate(light_e, oversample)
        elif oversample > 1:
            light_e = util.rebin(light_e, oversample)

        # shot noise
        if self.shot_noise:
            light_e = self.shot_noise(light_e)

        # prnu
        if self.prnu:
            light_e = self.prnu(light_e, window)

        # no negative electrons
        light_e[light_e < 0] = 0

        return np.floor(light_e)

    def dark(self, shape, ts, window):

        if ts > 0:
            # dark current
            if self.dark_signal:
                dark_flux = self.dark_signal(shape, window)
            else:
                dark_flux = np.zeros(shape)

            # integrate
            dark_e = self.integrate(dark_flux, ts)

            # apply dsnu
            if self.dsnu:
                if isinstance(self.dsnu, detector.FPN):
                    dark_e = self.dsnu(dark_e, window)
                elif isinstance(self.dsnu, (list, tuple)):
                    for dsnu in self.dsnu:
                        dark_e = dsnu(dark_e, window)
                else:
                    raise TypeError('DSNU must be an FPN object or list-like')

            # shot noise
            if self.dark_shot_noise:
                dark_e = self.dark_shot_noise(dark_e)

        else:
            dark_e = np.zeros(shape)

        # no negative electrons
        dark_e[dark_e < 0] = 0

        return np.floor(dark_e)

    def offset(self, electrons, window):
        # offset fpn
        if self.offset_fpn:
            if isinstance(self.offset_fpn, detector.FPN):
                electrons = self.offset_fpn(electrons, window)
            elif isinstance(self.offset_fpn, (list, tuple)):
                for fpn in self.offset_fpn:
                    electrons = fpn(electrons, window)
            else:
                raise TypeError('offset_fpn must be an OffsetFPN object or list-like')

        return electrons

    def readout(self, electrons, window, warn_saturate):
        # read noise
        if self.read_noise:
            electrons = self.read_noise(electrons)

        # no negative electrons
        electrons[electrons < 0] = 0

        # no partial electrons
        electrons = np.floor(electrons)

        # adc
        frame = self.gain(electrons, window, warn_saturate)

        # include focal plane defects
        if self.defects:
            if isinstance(self.defects, detector.Defect):
                frame = self.defects(frame, window)
            elif isinstance(self.defects, (list, tuple)):
                for defect in self.defects:
                    frame = defect(frame, window)

        return frame

    @staticmethod
    def integrate(img, ts):
        return np.floor(img * ts)

    @staticmethod
    def pixelate(img, oversample):
        pixel_sampling = Pixel()
        img = pixel_sampling(img, oversample)
        return util.rescale(img, 1/oversample, order=3, mode='nearest', unitary=True)


class BayerFPA(FPA):
    """Bayer focal plane array base class.

    """
    def __init__(self, pixelscale, shape, bayer_pattern, **kwargs):
        super().__init__(pixelscale, shape, **kwargs)
        self._bayer_pattern = bayer_pattern

        self._red_qe = kwargs.get('red_qe', None)
        self._green_qe = kwargs.get('green_qe', None)
        self._blue_qe = kwargs.get('blue_qe', None)

    @property
    def red_qe(self):
        """Red channel quantum efficiency

        Returns
        -------
        red_qe : :class:`~lentil.radiometry.Spectrum`

        """
        return self._red_qe

    @property
    def green_qe(self):
        """Green channel quantum efficiency

        Returns
        -------
        green_qe : :class:`~lentil.radiometry.Spectrum`

        """
        return self._green_qe

    @property
    def blue_qe(self):
        """Blue channel quantum efficiency

        Returns
        -------
        blue_qe : :class:`~lentil.radiometry.Spectrum`

        """
        return self._blue_qe

    @property
    def bayer_pattern(self):
        """A description of the detector's Bayer pattern. Each value represents
        the order of the red, green, and blue pixels starting with the upper-
        left corner of the image and moving left-to-right, top-to-bottom.` For
        example, ``'GBRG'`` represents the following Bayer pattern:

        = =
        G B
        R G
        = =

        Returns
        -------
        bayer_pattern : ``str``

        """
        if self._bayer_pattern:
            bayer_pattern = self._bayer_pattern.reshape(4)
            return (str(bayer_pattern[0]) + str(bayer_pattern[1]) +
                    str(bayer_pattern[2]) + str(bayer_pattern[3]))
        else:
            return None

    @bayer_pattern.setter
    def bayer_pattern(self, value):
        self._bayer_pattern = np.array(list(value.upper())).reshape(2, 2)

    @property
    def red_kernel(self):
        """Return red kernel"""
        return np.where(self._bayer_pattern == 'R', 1, 0)

    @property
    def blue_kernel(self):
        """Return blue kernel"""
        return np.where(self._bayer_pattern == 'B', 1, 0)

    @property
    def green_kernel(self):
        """Return green kernel"""
        return np.where(self._bayer_pattern == 'G', 1, 0)

    def collect_charge(self, count, wave, waveunit='nm', oversample=1, window=None):
        """Convert photon count (or flux) to electron count (or flux) by
        applying the detector's wavelength-dependent quantum efficiency.
        Additional processing is performed to apply separate QE curves and
        location masks for the separate red, green, and blue channels.

        Parameters
        ----------
        count : array_like
            The photons presented to the sensor. Should have shape (nwave,
            nrows, ncols)

        wave : array_like
            Wavelengths corresponding to each slice in ``count``. The length of
            ``wave`` must be equal to the first dimension in ``count``.

        waveunit : str, optional
            Units of ``wave``. Defaults is ``nm``

        oversample : int, optional
            Oversampling factor present in ``count``. Default is 1.

        window :  array_like or None, optional
            Indices of focal plane array to use given as (r_start, r_end,
            c_start, c_end). This definition follows standard Numpy indexing.

        Returns
        -------
        elec : ndarray
            Electron count or flux

        Notes
        -----
        * The units of ``count`` don't really matter, as long as the user is
          aware that this method converts photons per <whatever> to electrons
          per <whatever>. <Whatever> is nothing for counts and seconds for flux.
        * It is important to correctly specify :attr:`oversample` so that the
          correct pixel location mask is applied to any oversampled data.

        """
        nrow = count.shape[1]//oversample
        ncol = count.shape[2]//oversample

        # build up the bayer image. we do this one channel at a time, and then
        # accumulate the individual channels into one final frame
        red_qe = self.red_qe.sample(wave, waveunit=waveunit)
        red_mosaic = np.tile(self.red_kernel, (nrow//2, ncol//2))
        red_mosaic = ndimage.zoom(red_mosaic, oversample, order=0, mode='wrap')
        red_e = np.einsum('ijk,i->jk', count, red_qe) * red_mosaic

        green_qe = self.green_qe.sample(wave, waveunit=waveunit)
        green_mosaic = np.tile(self.green_kernel, (nrow//2, ncol//2))
        green_mosaic = ndimage.zoom(green_mosaic, oversample, order=0, mode='wrap')
        green_e = np.einsum('ijk,i->jk', count, green_qe) * green_mosaic

        blue_qe = self.blue_qe.sample(wave, waveunit=waveunit)
        blue_mosaic = np.tile(self.blue_kernel, (nrow//2, ncol//2))
        blue_mosaic = ndimage.zoom(blue_mosaic, oversample, order=0, mode='wrap')
        blue_e = np.einsum('ijk,i->jk', count, blue_qe) * blue_mosaic

        return red_e + green_e + blue_e



import warnings

import numpy as np
from scipy import signal

from lentil import util

__all__ = ['Gain', 'PolynomialGain', 'ShotNoise', 'GaussianShotNoise',
           'DarkCurrent', 'ReadNoise', 'Rule07DarkCurrent', 'FPN', 'RandomFPN',
           'NormalFPN', 'UniformFPN', 'LognormalFPN', 'PRNU', 'DarkCurrentFPN',
           'PixelOffsetFPN', 'ColumnOffsetFPN', 'ChargeDiffusion', 'Defect',
           'PixelMask', 'CosmicRay']


class Windowable:
    """Base class for representing data that can have a specific sub-window
    extracted.

    Parameters
    ----------
    data : float or array_like
        Windowable data.

    """

    def __init__(self, data):
        self.data = np.asarray(data)

    def window(self, shape=None, window=None):
        """Return an appropriately sized, potentially windowed :attr:`data`
        array.

        Parameters
        ----------
        shape : array_like or None, optional
            Output shape given as (nrows, ncols). If ``None`` (default), an
            uncropped array is returned.

        window : array_like or None, optional
            Indices of :attr:`data` array to return given as (r_start, r_end,
            c_start, c_end). This definition follows standard numpy indexing.

        Returns
        -------
        data : ndarray
            Trimmed and windowed :attr:`data` array

        Notes
        -----
        * If ``data`` is a single value (data.size == 1), self.data is returned
          regardless of what ``shape`` and ``window`` are.
        * If ``shape`` is given but ``window`` is ``None``, the returned ndarray
          is trimmed about the center of the array using
          :func:`lentil.util.pad`.
        * If ``window`` is given but ``shape`` is ``None``, the returned ndarray
          is extracted from :attr:`data` according to the indices in ``window``
        * If both ``shape`` and ``window`` are given, the returned ndarray is
          extracted from :attr:`data` according to the indices in ``window`` and
          the following expressions must be true:

        .. code::

            shape[0] = (window[1] - window[0]) = (r_end - r_start)
            shape[1] = (window[3] - window[2]) = (c_end - c_start)

        """
        if self.data.size == 1:
            return self.data

        if shape is None and window is None:
            # return the entire array
            return self.data

        elif window is not None:
            if shape is not None:
                # ensure consistency
                assert (window[1] - window[0]) == shape[0]
                assert (window[3] - window[2]) == shape[1]
            # return the windowed array.
            # Note that numpy implicitly handles a third dimension if one is
            # present
            return self.data[window[0]:window[1], window[2]:window[3]]

        else:  # shape is None
            # return the padded array
            # The same comment about numpy implicitly handling a third dimension
            # (see above) holds true here, although in this case the logic
            # resides in pad
            return util.pad(self.data, shape)


class Gain(Windowable):
    """Linear gain model.

    Parameters
    ----------
    gain : float or array_like
        Conversion gain in DN/e-. If specified as a single value, it is applied
        uniformly. If specified as an array, it is assumed to provide a
        pixel-by-pixel gain map.

    saturation_capacity : int or None
        Electron count resulting in pixel saturation. If None, pixels will not
        saturate. This is obviously nonphysical, but can be useful for testing
        or debugging.

    dtype : data-type or None, optional
        Output data-type. If None (default), no data-type casting is performed.

    Note
    ----
    The saturation capacity should not be confused with the full-well capacity.
    Saturation capacity is typically smaller than the full well capacity because
    the signal is clipped before the physical saturation of the pixel is
    reached.
    """

    def __init__(self, gain, saturation_capacity, dtype=None):
        super().__init__(gain)
        self.saturation_capacity = saturation_capacity
        self.dtype = dtype

    def __call__(self, img, window=None, warn_saturate=False):
        """Apply gain model to an array

        Parameters
        ---------
        img : ndarray
            Array of electron counts

        window : array_like or None, optional
            Indices of gain to return given as (r_start, r_end, c_start, c_end).
            This definition follows standard Numpy indexing.

        warn_saturate : bool, optional
            Raise a warning when pixels saturate. Default is False.

        Returns
        -------
        img : ndarray
            Array of DN

        """
        # enforce the saturation limit
        if self.saturation_capacity:

            if warn_saturate:
                if self.check_saturate(img):
                    warnings.warn('Frame has saturated pixels.')

            # Apply the saturation limit
            img[img > self.saturation_capacity] = self.saturation_capacity

        # convert to DN
        return np.floor(img * self.gain(img.shape, window), dtype=self.dtype)

    def gain(self, shape=None, window=None):
        """Return a trimmed and/or windowed representation of the gain model."""
        return self.window(shape, window)

    def check_saturate(self, img):
        if np.any(img > self.saturation_capacity):
            return True
        else:
            return False


class PolynomialGain(Gain):
    """Gain model defined by polynomial coefficients.

    Parameters
    ----------
    gain : array_like
        Conversion gain in DN/e-. Can be specified in two ways:

            * As a one-dimensional array of polynomial coefficients applied
              globally to each pixel
            * As a three-dimensional array of pixel-by-pixel gain where the
              first dimension gives polynomial coefficients of each pixel

    saturation_capacity : int or None
        Electron count resulting in pixel saturation. If None, pixels will not
        saturate. This is obviously nonphysical, but can be useful for testing
        or debugging.

    dtype : data-type or None, optional
        Output data-type. If None (default), no data-type casting is performed.

    Note
    ----
    The saturation capacity should not be confused with the full-well capacity.
    Saturation capacity is typically smaller than the full well capacity because
    the signal is clipped before the physical saturation of the pixel is
    reached.

    """

    def __init__(self, gain, saturation_capacity, dtype=None):

        gain = np.asarray(gain)
        if not np.any([gain.ndim == 1, gain.ndim == 3]):
            raise ValueError

        super().__init__(gain, saturation_capacity, dtype)

    def __call__(self, img, window=None, warn_saturate=False):
        """Apply gain model to an array

        Parameters
        ---------
        img : ndarray
            Array of electron counts

        window : array_like or None, optional
            Indices of gain to return given as (r_start, r_end, c_start, c_end).
            This definition follows standard Numpy indexing.

        warn_saturate : bool, optional
            Raise a warning when pixels saturate. Default is False.

        Returns
        -------
        img : ndarray
            Array of DN

        """
        # enforce the saturation limit
        if self.saturation_capacity:

            if warn_saturate:
                if self.check_saturate(img):
                    warnings.warn('Frame has saturated pixels.')

            # Apply the saturation limit
            img[img > self.saturation_capacity] = self.saturation_capacity

        # Determine the polynomial order
        if self.data.ndim == 1:
            model_order = self.data.shape[0]
        else:
            model_order = self.data.shape[2]

        # Prepare a cube of electron counts to apply the polynomial gain to
        img_cube = np.repeat(img[:, :, np.newaxis], model_order-1, axis=2)
        for order in np.arange(model_order-1, 0, -1):
            d = model_order - order - 1
            img_cube[..., d] = img_cube[..., d]**order

        # Get the appropriately sized and windowed gain model
        gain_model = self.gain(img.shape, window)

        # Apply the gain model and convert to DN
        if self.data.ndim == 1:
            gain, offset = gain_model[0:-1], gain_model[-1]
            return np.einsum('ijk,k->ij', img_cube, gain) + offset
        else:
            # self.data.ndim == 3
            gain, offset = gain_model[..., 0:-1], gain_model[..., -1]
            return np.einsum('ijk,ijk->ij', img_cube, gain) + offset

    def gain(self, shape=None, window=None):
        """Return a trimmed and/or windowed representation of the gain model."""
        if self.data.ndim == 1:
            # we have a global nonlinear gain where the shape and window don't
            # matter
            return self.data
        else:
            # self.data.ndim == 3
            return self.window(shape, window)


class RandomNoise:
    """Base class for all radiometry of random noise with an optional
    user-definable seed.

    If a seed is provided, an instance of ``numpy.random.RandomState`` is
    created and all random numbers are generated from this object in a
    repeatable (but still random) way. If no seed is provided, random numbers
    are generated using the appropriate ``numpy.random`` distribution.

    Parameters
    ----------
    seed : None, int, or array_like, optional
        Random seed used to initialize ``numpy.random.RandomState``. If
        ``None``, no random number generator object is created.

    Attributes
    ----------
    rng : numpy.random.RandomState or None
        Random number generator object. If ``seed`` is ``None``, ``rng`` will
        also be ``None``
    """

    def __init__(self, seed=None):
        self.seed = seed

        if seed:
            self.rng = np.random.RandomState(seed)
        else:
            self.rng = None


class ShotNoise(RandomNoise):
    r"""Shot noise modeled by a Poisson process.

    Parameters
    ----------
    seed : None, int, or array_like, optional
        Seed for random number generator. If ``None`` (default), the Numpy
        default is used.

    Note
    ----
    For sufficiently large values of :math:`\lambda`,  (say :math:`\lambda >
    1000`), the Normal(:math:`\mu = \lambda`, :math:`\sigma^2 = \lambda`)
    distribution is an excellent approximation to the Poisson(:math:`\lambda`)
    distribution and is about 25% faster to compute.

    See Also
    --------
    :class:`~lentil.detector.GaussianShotNoise` Shot noise modeled by a normal
    distribution

    """

    def __init__(self, seed=None):
        super().__init__(seed)

    def __call__(self, img):
        """Apply shot noise model to an array

        Parameters
        ---------
        img : ndarray
            Array of electron counts

        Returns
        -------
        img : ndarray
            Array of noisy electron counts

        """
        if self.rng:
            return self.rng.poisson(img)
        else:
            return np.random.poisson(img)


class GaussianShotNoise(RandomNoise):
    r"""Shot noise modeled by a normal distribution.

    Parameters
    ----------
    seed : None, int, or array_like, optional
        Seed for random number generator. If ``None`` (default), the Numpy
        default is used.

    Note
    ----
    For sufficiently large values of :math:`\lambda`,  (say :math:`\lambda >
    1000`), the Normal(:math:`\mu = \lambda`, :math:`\sigma^2 = \lambda`)
    distribution is an excellent approximation to the Poisson(:math:`\lambda`)
    distribution and is about 25% faster to compute.

    See Also
    --------
    :class:`~lentil.detector.ShotNoise` Shot noise modeled by a Poisson process

    """

    def __init__(self, seed=None):
        super().__init__(seed)

    def __call__(self, img):
        """Apply shot noise model to an array

        Parameters
        ---------
        img : ndarray
            Array of electron counts

        Returns
        -------
        img : ndarray
            Array of noisy electron counts

        """
        if self.rng:
            return self.rng.normal(loc=img, scale=img)
        else:
            return np.random.normal(loc=img, scale=img)


class ReadNoise(RandomNoise):
    """Read noise model.

    Parameters
    ----------
    electrons : int
        Read noise per frame

    seed : None, int, or array_like, optional
        Seed for random number generator. If ``None`` (default), the Numpy
        default is used.

    """
    def __init__(self, electrons, seed=None):
        super().__init__(seed)
        self.electrons = electrons

    def __call__(self, frame):
        if self.rng:
            noise = self.rng.normal(loc=0.0, scale=self.electrons, size=frame.shape)
        else:
            noise = np.random.normal(loc=0.0, scale=self.electrons, size=frame.shape)
        return frame + noise


class DarkSignal(Windowable):
    """Base class for representing a dark signal.

    Parameters
    ----------
    value : float, or array_like
        Dark signal value. If specified as a single value, it is applied
        uniformly. If specified as an array, it is assumed to provide a
        pixel-by-pixel map.

    """
    def __init__(self, value):
        super().__init__(value)

    def __call__(self, shape, window=None):
        """Create dark signal array

        Parameters
        ---------
        shape : array_like
            Dimensions of dark signal array to create

        window : array_like or None, optional
            Indices of dark signal to return given as (r_start, r_end, c_start,
            c_end). This definition follows standard Numpy indexing.

        Returns
        -------
        dark_signal : ndarray
            Dark signal array

        """
        # Return ones * the dark() value in the event that a single value is
        # provided (and thus returned by dark)
        dark_signal = self.dark(shape, window) * np.ones(shape)

        return dark_signal

    def dark(self, shape, window):
        return self.window(shape, window)


class DarkCurrent(DarkSignal):
    """Generic dark current model.

    Parameters
    ----------
    rate : float or array_like
        Dark current rate in electrons/sec/px. If specified as a single value,
        it is applied uniformly. If specified as an array, it is assumed to
        provide a pixel-by-pixel map.

    See Also
    --------
    :class:`~lentil.detector.Rule07DarkCurrent` Rule 07 dark current model for
    IR detectors

    """

    def __init__(self, rate):
        super().__init__(rate)

    @property
    def rate(self):
        return self.data


class Rule07DarkCurrent(DarkCurrent):
    """Model for representing dark current in HgCdTe infrared detectors using
    Rule 07.

    Parameters
    ----------
    temperature : float
        Focal plane temperature in K

    cutoff_wavelength : float
        Focal plane cutoff wavelength in m

    pixelscale : float
        Size of one pixel in m

    References
    ----------
    * Tennant, W.E. et al. MBE HgCdTe Technology: A Very General Solution to IR
      Detection, Described by "Rule 07", a Very Convenient Heuristic. Journal of
      Electronic Materials (2008).

    """

    def __init__(self, temperature, cutoff_wavelength, pixelscale):

        self.temperature = temperature
        self.cutoff_wavelength = cutoff_wavelength

        J0 = 8367.00001853855  # A/cm^2
        C = -1.16239134096245
        k = 1.3802e-23  # J/K - Boltzmanns' constant
        q = 1.6021e-19  # C - electron charge

        lambda_threshold = 4.63513642316149
        lambda_scale = 0.200847413564122
        P = 0.544071281108481

        # compute lambda_e
        lambda_cutoff = cutoff_wavelength * 1e6  # meters to microns
        if lambda_cutoff >= lambda_threshold:
            lambda_e = lambda_cutoff
        else:
            lambda_e = lambda_cutoff/(1-((lambda_scale/lambda_cutoff) -
                                         (lambda_scale/lambda_threshold))**P)

        # apply Rule 07 to compute the dark flux in A/cm^2
        J = J0*np.exp(C*(1.24*q/(k*lambda_e*temperature)))

        # convert J in A/cm^2 to J in e-/px/s
        px_area = (pixelscale*1e2)**2  # pixel area in cm^2
        super().__init__(rate=1/q * px_area * J)

    def __call__(self, shape, *args, **kwargs):
        """Create dark signal array

        Parameters
        ---------
        shape : array_like
            Dimensions of dark signal array to create

        Returns
        -------
        dark_signal : ndarray
            Dark signal array

        """
        return self.rate * np.ones(shape)


class FPN(Windowable):
    """Base class for all radiometry of fixed-pattern noise .

    Parameters
    ----------
    data : array_like
        FPN array.

    """
    def __init__(self, data):
        super().__init__(data)

    def __call__(self, frame, window=None):
        """Apply fixed pattern noise to an array.

            The output array :math:`A'` is related to the input array :math:`A`
            by

            .. math::

                A' = A*F

            where :math:`F` is the :attr:`data` FPN array.

        Parameters
        ----------
        frame : ndarray

        window : array_like or None, optional
            Indices of :attr:`data` FPN array to return given as (r_start,
            r_end, c_start, c_end). This definition follows standard Numpy
            indexing.

        Returns
        -------
        frame : ndarray
            Array with  FPN applied

        """
        return frame * self.fpn(frame.shape, window)

    def fpn(self, shape=None, window=None):
        return self.window(shape, window)


class RandomFPN(FPN):
    """Base class for randomly generated fixed-pattern noise with an optional
    user-definable seed.

    Parameters
    ----------
    seed : None, int, or array_like, optional
        Seed for random number generator. If ``None`` (default), the Numpy
        default is used.

    Attributes
    ----------
    rng : numpy.random.RandomState or None
        Random number generator object

    data : ndarray
        FPN array.

    """
    def __init__(self, seed=None):
        self.seed = seed
        if seed:
            self.rng = np.random.RandomState(seed)
        else:
            self.rng = None

        self.data = None


class NormalFPN(RandomFPN):
    """Base class for fixed pattern noise represented by a normal distribution.

    Parameters
    ----------
    loc : float

    scale : float

    size : float

    """
    def __init__(self, loc, scale, size, seed=None):
        super().__init__(seed)
        if self.rng:
            self.data = self.rng.normal(loc, scale, size)
        else:
            self.data = np.random.normal(loc, scale, size)


class LognormalFPN(RandomFPN):
    """Base class for fixed pattern noise represented by a log-normal
    distribution.

    Parameters
    ----------
    mean : float

    sigma : float

    size : float

    """
    def __init__(self, mean, sigma, size, seed=None):
        super().__init__(seed)
        if self.rng:
            self.data = self.rng.lognormal(mean, sigma, size)
        else:
            self.data = np.random.lognormal(mean, sigma, size)


class UniformFPN(RandomFPN):
    """Base class for fixed pattern noise represented by a uniform distribution.

    Parameters
    ----------
    low : float

    high : float

    size : float

    """
    def __init__(self, low, high, size, seed=None):
        super().__init__(seed)
        if self.rng:
            self.data = self.rng.uniform(low, high, size)
        else:
            self.data = np.random.uniform(low, high, size)


class PRNU(NormalFPN):
    """Model for representing photoresponse non-uniformity.

    """
    def __init__(self, prnu_factor, size, seed=None):
        self.prnu_factor = prnu_factor
        super().__init__(loc=1.0, scale=prnu_factor, size=size, seed=seed)


class DarkCurrentFPN(LognormalFPN):
    """Dark current FPN model, represented by a log-normal distribution.

    Parameters
    ----------
    sigma : float
        Dark current FPN factor, usually between 0.1 and 0.4 for CCD and CMOS
        detectors [2].

    size : None, tuple, or array_like, optional
        Precomputed FPN matrix size. Typically this will be equal to the
        detector size. See :class:`GainFPN` note for details on the two
        different modes for when ``size`` is specified or not.

    seed : None, int, or array_like, optional
        Random seed used to initialize ``numpy.random.RandomState``. If
        ``None``, then ``RandomState`` will try to read data from /dev/urandom
        (or the Windows analogue) if available or seed from the clock otherwise.

    References
    ----------
    [1] `Log-normal distribution - Wikipedia <https://en.wikipedia.org/wiki/Log-normal_distribution>`_

    [2] Janesick, J. R., Photon Transfer, Vol. PM170, SPIE (2007)

    """

    def __init__(self, sigma, size, seed=None):
        super().__init__(mean=1.0, sigma=sigma, size=size, seed=seed)
        self.sigma = sigma


class PixelOffsetFPN(NormalFPN):
    r"""Pixel offset FPN model.

    Pixel offset FPN can be represented by an autoregressive random process with
    (:math:`\mu = 0.0, \sigma = ` :attr:`sigma`) [1], [2].

    Parameters
    ----------
    sigma : float
        1-sigma pixel offset FPN value in number of electrons

    ar_parameter : float
        Autoregressive parameter which characterizes the dependency of each
        pixel on its neighbors. Valid values are between 0 and 0.5 [1]. Small
        values correspond to less correlation.

    size : None or array_like, optional
        Precomputed FPN matrix size. Typically this will be equal to the
        detector size. See :class:`GainFPN` note for details on the two
        different modes for when ``size`` is specified or not.

    seed : None, int, or array_like, optional
        Random seed used to initialize ``numpy.random.RandomState``. If
        ``None``, then ``RandomState`` will try to read data from /dev/urandom
        (or the Windows analogue) if available or seed from the clock otherwise.

    References
    ----------
    [1] El Gamal, A. et. al., Modeling and Estimation of FPN Components in CMOS Image Sensors, SPIE Proc. 3301 (1998)

    [2] `Autoregressive model - Wikipedia <https://en.wikipedia.org/wiki/Autoregressive_model>`_

    """

    def __init__(self, sigma, ar_parameter, size, seed=None):
        assert (ar_parameter >= 0) and (ar_parameter <= 0.5)

        super().__init__(loc=1.0, scale=sigma, size=size, seed=seed)
        self.sigma = sigma
        self.ar_parameter = ar_parameter

        # apply the autoregressive filter
        self.data = signal.lfilter([1], [1, self.ar_parameter], self.data, axis=0)


class ColumnOffsetFPN(NormalFPN):
    r"""Column offset FPN model.

    Column offset FPN can be represented by an autoregressive random process
    with (:math:`\mu = 0.0, \sigma = ` :attr:`sigma`) [1], [2].

    Parameters
    ----------
    sigma : float
        1-sigma column offset FPN value in number of electrons

    ar_parameter : float
        Autoregressive parameter which characterizes the dependency of each
        column on its neighbors. Valid values are between 0 and 0.5 [1]. Small
        values correspond to less correlation.

    size : None or array_like, optional
        Precomputed FPN matrix size. Typically this will be equal to the
        detector size. See :class:`GainFPN` note for details on the two
        different modes for when ``size`` is specified or not.

    seed : None, int, or array_like, optional
        Random seed used to initialize ``numpy.random.RandomState``. If
        ``None``, then ``RandomState`` will try to read data from /dev/urandom
        (or the Windows analogue) if available or seed from the clock otherwise.

    References
    ----------
    [1] El Gamal, A. et. al., Modeling and Estimation of FPN Components in CMOS
        Image Sensors, SPIE Proc. 3301 (1998)

    [2] `Autoregressive model - Wikipedia <https://en.wikipedia.org/wiki/Autoregressive_model>`_

    """

    def __init__(self, sigma, ar_parameter, size, seed=None):
        assert (ar_parameter >= 0) and (ar_parameter <= 0.5)

        super().__init__(loc=1.0, scale=sigma, size=size[1], seed=seed)
        self.sigma = sigma
        self.ar_parameter = ar_parameter

        # apply the autoregressive filter
        self.data = signal.lfilter([1], [1, self.ar_parameter], self.data)
        self.data = np.tile(self.data, (size[0], 1))


class ChargeDiffusion:
    """Charge diffusion model represented by a gaussian blur."""
    def __init__(self, sigma):
        self.sigma = sigma
        self.kernel = util.gaussian2d(3, self.sigma)

    def __call__(self, img):
        return signal.convolve2d(img, self.kernel, mode='same')


class Defect(Windowable):
    """Base class for representing detector defects."""
    pass


class PixelMask(Defect):
    """Class for representing defective, nonlinear, or dead pixels by providing
    a mask.
    """
    pass


class CosmicRay:
    """Cosmic ray generator for simulating cosmic ray hits on a detector.

    The default values for cosmic ray rates and electron fluxes are taken from
    [1]. It is commonly accepted at about 90% of cosmic rays are protons and the
    majority of the remaining 10% are alpha particles [1], [2].

    Parameters
    ----------
    rate : float, optional
        Cosmic ray rate expressed as number of events per square meter of
        detector area. Default is 40000 hits/m^2.

    proton_flux : float, optional
        Number of electrons liberated per meter of travel for a proton. Default
        is 1e9 e-/m.

    alpha_flux : float, optional
        Number of electrons liberated per meter of travel for an alpha particle.
        By definition, alpha particles have four times the energy of protons.
        Default is 4e9 e-/m.

    Example
    -------
    Simulate the cosmic ray hits for a 300 second exposure over a 256 x 256 detector
    patch with 5 um x 5 um x 3 um pixels:

    .. code:: pycon

        >>> import matplotlib.pyplot as plt
        >>> import lentil
        >>> cosmic_generator = lentil.detector.CosmicRay()
        >>> cosmic_frame = cosmic_generator((256,256), (5e-6, 5e-6, 3e-6), 300)
        >>> plt.imshow(cosmic_frame)

    .. image:: ../_static/img/api/detector/cosmic_ray.png
        :width: 300px

    References
    ----------
    [1] Offenberg, J.D. et. al.  Multi-Read Out Data Simulator. (2000).

    [2] `Cosmic ray - Wikipedia <https://en.wikipedia.org/wiki/Cosmic_ray>`_
    """

    def __init__(self, rate=4e4, proton_flux=1e9, alpha_flux=4e9):
        self.rate = rate
        self.proton_flux = proton_flux
        self.alpha_flux = alpha_flux

    def __call__(self, shape, pixelscale, ts):
        """Simulate cosmic ray hits on a detector.

        Parameters
        ----------
        shape : array_like
            Frame size defined as (rows, cols)

        pixelscale : array_like
            Pixel dimensions as (y,x,z) where z is the pixel depth

        ts : float
            Integration time in seconds

        Returns
        -------
        cosmic_e : ndarray
            Array representing the number of electrons in a detector image due
            to cosmic ray hits


        """

        # compute the number of rays that strike the detector during the
        # integration time
        nrays = self.nrays(shape, pixelscale, ts)

        img = np.zeros(shape)
        for ray in np.arange(0, nrays):
            img += self.cosmic_ray(shape, pixelscale)
        return img

    def nrays(self, shape, pixelscale, ts):
        area = shape[0]*pixelscale[0]*shape[1]*pixelscale[1]

        nrays = area * self.rate * ts

        # even if we don't have any events as defined by the rate, there is
        # still a chance we will see an event
        if nrays < 1:
            if np.random.uniform() <= nrays:
                nrays = 1
            else:
                nrays = 0

        return int(nrays)

    def cosmic_ray(self, shape, pixelscale):

        # randomly choose which type of particle we have
        if np.random.uniform() > 0.9:
            # alpha particle
            electron_flux = self.alpha_flux
        else:
            # proton
            electron_flux = self.proton_flux

        img = np.zeros(shape)

        # random position
        r = np.random.rand() * (shape[0]-1)
        c = np.random.rand() * (shape[1]-1)
        z = 0
        position = np.array([r, c, z])

        # give the cosmic ray a random direction (theta) and angle of incidence
        # (phi) and compute a unit vector representing the ray's direction of
        # travel. we can make things slightly easier on ourselves by limiting
        # phi to the interval 0 < phi < pi. this is assumed to be valid because
        # a cosmic ray striking the detector from behind will have essentially
        # the same effect as one striking it from the front. phi is negative to
        # force the z-direction of v to be negative (traveling into the detector
        # from above)
        theta = np.random.uniform() * 2 * np.pi
        phi = np.random.uniform() * 1*np.pi
        direction = np.array([np.cos(theta)*np.cos(phi), np.sin(theta)*np.cos(phi),
                              -np.sin(phi)])

        # scale the direction vector so that we can use ray_box_exit() assuming
        # a cube instead of a potentially rectangular pixel size
        pixelscale = np.asarray(pixelscale)
        scale = pixelscale/np.max(pixelscale)
        direction /= scale
        direction /= np.linalg.norm(direction)

        extent = (0, shape[0]-1, 0, shape[1]-1, 0, -1)

        # propagate the ray
        ray = self.propagate(position, direction, extent)

        for i in np.arange(0, ray.shape[0]-1):

            # compute the distance the ray traveled in pixel space
            dr = (ray[i+1][0]-ray[i][0])*pixelscale[0]
            dc = (ray[i+1][1]-ray[i][1])*pixelscale[1]
            dz = (ray[i+1][2]-ray[i][2])*pixelscale[2]
            dist = np.sqrt(dr**2+dc**2+dz**2)

            row = int(np.floor(ray[i+1][0]))
            col = int(np.floor(ray[i+1][1]))
            img[row, col] += electron_flux*dist

        return img

    @staticmethod
    def propagate(position, direction, extent):
        position = position[np.newaxis, ...]
        direction = direction[np.newaxis, ...]

        intersections = cubeplane_ray_intersection(position, direction, extent)
        raw_ray_data = process_cube_intersections(intersections, position, direction)

        return raw_ray_data[3]


def cubeplane_ray_intersection(xyz, dxdydz, extent):
    """Intersects rays with a plane of cubes and returns data about
    intersections

    Suppose the walls of the cube are on integer boundaries in the space of
    extent[0] <= x <= extent[1], extent[2] <= y <= extent[3], extent[4] <= z <=
    extent[5]

    Parameters
    ----------
    xyz : ndarray, ndim=2
        xyz[:, 0] is the x coordinate of the starting point of the ray,
        xyz[:, 1] is the y coordinate of the starting point of the ray.
        xyz[:, 2] is the z coordinate of the starting point.

    dxdydz : double ndim=2
        Unit vector representing the direction of travel of the ray.

    extent : tuple of int
        Integer address for the detector edge on the -x, +x, -y,
        +y, -z and +z edges. Should be representable as an int16.

    Returns
    -------
    inter : structured ndarray
        This is an array of structures with the dtype ('raylength', 'f4'),
        ('indx', 'u8'), ('facedirection', 'u1') where indx is the number of the
        row for the ray in xyz/dxdydz, raylength is the distance along the ray
        of the intersection and facedirection is an indicator of the ray
        direction. If face_direction = 0, 2, 4, the ray passed through the
        constant X, Y or Z planes, respectively, going from positive to
        negative. If face_direction = 1, 3, 5, the ray was going from negative
        to positive X, Y or Z through a surface.

    Notes
    -----
    If you don't like the output of this function, feed it into
    process_cube_intersections().

    This code has been tested when the rays start outside the cube volume. Would
    recommend not starting rays inside cube or within 2e-7 of a cube boundary.

    Very very very slightly adapted from code kindly developed by Dustin Moore.

    """

    xyz = xyz.astype(np.float32)
    dxdydz = dxdydz.astype(np.float32)

    indx = np.array(range(xyz.shape[0]), dtype=np.uint64)

    extent_shaped = np.array(extent, dtype=np.float32).reshape((1, 3, 2))

    # CUBE INTERSECTION
    # distance along ray before it penetrates outer CCD grid on xyz, -/+ side
    max_Ls = np.true_divide(extent_shaped - xyz[:, :, None], dxdydz[:, :, None])
    max_Ls[np.isinf(max_Ls)] = np.nan

    # Find length of ray to point where it enters cube array
    min_L = np.nanmax(np.min(max_Ls, axis=2), axis=1)

    # Find total length of ray to point as it exits cube array
    max_L = np.nanmin(np.max(max_Ls, axis=2), axis=1)
    del max_Ls

    # Find rays that never intersect cube or only intersect if the ray traced
    # backward
    no_intersection_mask = np.logical_or(max_L < min_L, max_L <= 0.0)

    # ... and chuck them from the calculation
    if np.any(no_intersection_mask):
        keep_mask = np.invert(no_intersection_mask)
        xyz = xyz[keep_mask, :]
        dxdydz = dxdydz[keep_mask, :]
        indx = indx[keep_mask]
        min_L = min_L[keep_mask]
        max_L = max_L[keep_mask]
        del keep_mask
    del no_intersection_mask

    # CUBE EDGE INTERSECTIONS
    # Shortest travel along each axis before ray exits cube array bounding
    # extent
    min_L[min_L < 0.0] = 0.0  # Rays can not go backward
    nearest = xyz + (min_L[:, None] - 1e-5) * dxdydz
    del min_L
    # Furthest travel along each axis before ray exits cube array bounding
    # extent
    furthest = xyz + (max_L[:, None] + 1e-5) * dxdydz
    del max_L

    d_pos = dxdydz >= 0.0
    least = np.ceil(np.where(d_pos, nearest, furthest)).astype(np.int16)
    great = np.floor(np.where(d_pos, furthest, nearest)).astype(np.int16) + 1

    d_pos = d_pos.astype(np.uint8)
    d_pos[:, 1] += 2
    d_pos[:, 2] += 4

    # No iterate when intersection impossible
    np.putmask(great, dxdydz == 0.0, -32768)

    intersections = []  # O(1) appends for the win.

    for xy, dxd, lea, grea, id, dp in zip(xyz, dxdydz, least, great, indx, d_pos):
        # algorithm-limiting performance here
        this_ints = [((q - xy[0]) / dxd[0], id, dp[0]) for q in range(lea[0], grea[0])]

        this_ints += [((q - xy[1]) / dxd[1], id, dp[1]) for q in range(lea[1], grea[1])]

        this_ints += [((q - xy[2]) / dxd[2], id, dp[2]) for q in range(lea[2], grea[2])]

        intersections += sorted(this_ints)

    # We repack the list into a numpy object for ease of conveyence back to the
    # mother process and later numpy computations
    fancy_dtype = np.dtype([('raylength', 'f4'), ('indx', 'u8'),
                            ('facedirection', 'u1')], align=True)
    inter = np.array(intersections, dtype=fancy_dtype)

    return inter


def process_cube_intersections(inter, xyz, dxdydz, verbose=False):
    """Intersects rays with a plane of cubes and returns data about
    intersections

    Suppose the walls of the cube are on integer boundaries in the space of
    extent[0] <= x <= extent[1], extent[2] <= y <= extent[3], extent[4] <= z
    <= extent[5]

    Parameters
    ----------
    inter : list of tuple
        Output of cubeplane_ray_intersection() or _cubeplane_ray_intersection()

    xyz : ndarray, ndim=2
        xyz[:, 0] is the x coordinate of the starting point of the ray,
        xyz[:, 1] is the y coordinate of the starting point of the ray.
        xyz[:, 2] is the z coordinate of the starting point.

    dxdydz : double ndim=2
        Unit vector representing the direction of travel of the ray.

    verbose: bool, optional
        If True, prints every ray intersection in human readable format. Default
        is False.

    Returns
    -------
    indx : ndarray of np.uint64
        Index of the ray for each ray intersection

    raylength : ndarray of np.float32
        Length along each ray that the intersection happened

    draylength : ndarray of np.float32
        Length along ray travsersed since ray was born or last intersection,
        which ever came most recently

    interpoint : ndarray of np.float32
        One row for each ray intersection, values are x,y,z of intersection
        point

    last_coord : ndarray of np.int16
        One row for each ray intersection, values are the x,y,z coordinate of
        the cube that the ray is leaving. (May not be inside extent, you should
        check if you care.)

    next_coord : ndarray of np.int16
        One row for each ray intersection, values are the x,y,z coordinate of
        the cube that the ray is entering. (May not be inside extent, you should
        check if you care.)

    Notes
    -----
    This code has been tested when the rays start outside the cube
    volume. Would recommend not starting rays inside cube or within 2e-7 of a
    cube boundary.

    Originally developed by Dustin Moore.

    """

    _xyz = xyz[inter['indx'], :].astype(np.float32)
    _dxdydz = dxdydz[inter['indx'], :].astype(np.float32)

    draylength = np.ediff1d(inter['raylength'], to_begin=(0.0,))
    np.putmask(draylength, np.ediff1d(inter['indx'], to_begin=(1,)) != 0,
               inter['raylength'])

    inter_point = _xyz + inter['raylength'][:, None] * _dxdydz

    next_pixel_bias = np.array([[-1e-2, 0.0, 0.0], [1e-2, 0.0, 0.0],
                                [0.0, -1e-2, 0.0], [0.0, 1e-2, 0.0],
                                [0.0, 0.0, -1e-2],  [0.0, 0.0, 1e-2]])

    w = np.choose(inter['facedirection'][:, None], next_pixel_bias)
    next_coord = np.empty(_xyz.shape, dtype=np.int16)
    np.floor(inter_point + w, out=next_coord, casting='unsafe')

    last_coord = np.empty(_xyz.shape, dtype=np.int16)
    np.floor(inter_point - w, out=last_coord, casting='unsafe')

    if verbose:
        last_indx = -1
        for ind, rl, ip, lc, nc, dl in zip(inter['indx'], inter['raylength'],
                                           inter_point, last_coord, next_coord,
                                           draylength):
            if last_indx != ind:
                print('Ray ' + str(ind) + ' is born at ' + str(xyz[ind, :]) +
                      ' and heads off in direction ' + str(dxdydz[ind, :]))
            print('  then leaves cube ' + str(lc) + ' at ' + str(ip) +
                  ' for cube ' + str(nc) + ' after having gone  ' +
                  str(dl) + ' and traveling ' + str(rl) + ' from inception')
            last_indx = ind

    return (inter['indx'], inter['raylength'], draylength, inter_point,
            last_coord, next_coord)
